
library IEEE;
use IEEE.std_logic_1164.all;
USE ieee.std_logic_unsigned.ALL; 


entity ControlUnit is 
  generic (BitWidth: integer;
           InstructionWidth: integer);
  port(
    ----------------------------------------    
    rst: in std_logic;
    clk: in std_logic;     
    ----------------------------------------    
    Instr_In: in std_logic_vector (InstructionWidth-1 downto 0);
    Instr_Add: out std_logic_vector (BitWidth-1 downto 0);    
    ----------------------------------------    
    MemRdAddress: out std_logic_vector (BitWidth-1 downto 0); 
	 MemWrtAddress: out std_logic_vector (BitWidth-1 downto 0);
    Mem_RW: out std_logic;
    ----------------------------------------      DPU Signals
    DPU_Flags: in std_logic_vector (3 downto 0);
    DataToDPU: out std_logic_vector (BitWidth-1 downto 0);
    CommandToDPU: out std_logic_vector (10 downto 0);
	 Reg_in_sel: out std_logic_vector (7 downto 0);
	 Reg_out_sel: out std_logic_vector (2 downto 0); 
    DataFromDPU: in std_logic_vector (BitWidth-1 downto 0)
    
    ----------------------------------------
  );
end ControlUnit;


architecture RTL of ControlUnit is
  
  ---------------------------------------------
  --      Signals and Types
  ---------------------------------------------

  TYPE STATE_TYPE IS (Fetch, Decode, Execution,WriteBack);
  signal State_in, State_out :STATE_TYPE;
  
  TYPE Instruction IS (PUSH,POP,
                       JMPEQ,Jmp_rel,Jmp,JmpZ,JmpOV,JmpC,
                       FlipA,And_A_R,OR_A_R,XOR_A_R,NegA,
                       ShiftA_R,ShiftA_L,ShiftArithL,ShiftArithR,
                       RRC,RLC,
                       LoadPC,SavePC,
                       Add_A_R, Add_A_Mem,Add_A_Dir, Sub_A_R,Sub_A_Mem,Sub_A_Dir,IncA,DecA,
                       Load_A_Mem,Load_R0_Mem,Load_R0_Dir,Store_A_Mem,load_A_R,load_R_A,Load_Ind_A,
                       ClearZ,ClearOV,ClearC, ClearACC,
                       NOP,HALT);
  signal Instr,Instr0,Instr1,Instr2,Instr3,InstrForExecute,InstrForWriteBack:Instruction;  
  signal Instr0_in,Instr1_in,Instr2_in,Instr3_in,InstrForExecute_in,InstrForWriteBack_in:Instruction;
  
  signal SP_in, SP_out : std_logic_vector (BitWidth-1 downto 0):= (others => '0');
  signal PC_in, PC_out : std_logic_vector (BitWidth-1 downto 0):= (others => '0');
  signal InstrReg_out,InstrReg_out2,InstrReg_out3,InstrReg_out4: std_logic_vector (InstructionWidth-1 downto 0);
  signal State_out3,State_out2, State_out1 :STATE_TYPE;
  signal inst_counter: std_logic_vector (1 downto 0);
  
  signal DataFromDPU0,DataFromDPU1,DataFromDPU2,DataFromDPU3: std_logic_vector (BitWidth-1 downto 0);
  signal DataFromDPU0_in,DataFromDPU1_in,DataFromDPU2_in,DataFromDPU3_in: std_logic_vector (BitWidth-1 downto 0);
  signal DPU_Flags0,DPU_Flags1,DPU_Flags2,DPU_Flags3: std_logic_vector (3 downto 0);
   
  ---------------------------------------------
  --      OpCode Aliases
  ---------------------------------------------
  Alias opcode0 : std_logic_vector (5 downto 0) is InstrReg_out (InstructionWidth-1 downto BitWidth); 
  Alias opcode1 : std_logic_vector (5 downto 0) is InstrReg_out2 (InstructionWidth-1 downto BitWidth); 
  Alias opcode2 : std_logic_vector (5 downto 0) is InstrReg_out3 (InstructionWidth-1 downto BitWidth); 
  Alias opcode3 : std_logic_vector (5 downto 0) is InstrReg_out4 (InstructionWidth-1 downto BitWidth); 
  
  signal opcode : std_logic_vector (5 downto 0);
  signal Flush: std_logic;
  begin
  
  ---------------------------------------------
  -- Registers setting  
  ---------------------------------------------
  process (clk,rst)
    begin
    if rst = '1' then 
        
       State_out <= Fetch; 
       SP_out <= (others => '0');         
       PC_out <= (others => '0');
       InstrReg_out <= (others => '0');
       
       inst_counter <= (others => '0');
       
       InstrReg_out2<= (others => '0');
       InstrReg_out3<= (others => '0');
       InstrReg_out4<= (others => '0');
       
       State_out1 <= Fetch;
       State_out2 <= Fetch;
       State_out3 <= Fetch;
       
       Instr0 <= NOP;
       Instr1 <= NOP;
       Instr2 <= NOP;
       Instr3 <= NOP;
       
        DataFromDPU0 <= (others=> '0');
 		DataFromDPU1 <= (others=> '0'); 
 		DataFromDPU2 <= (others=> '0');
 		DataFromDPU3 <= (others=> '0');
 		 
 		
    elsif clk'event and clk='1' then
    	State_out <= State_in;
       	SP_out <= SP_in;
       	PC_out <= PC_in;
       	 
       	inst_counter <= inst_counter +1;
       	
       	InstrForExecute <=InstrForExecute_in;
       	Instr0 <= Instr0_in;
       	Instr1 <= Instr1_in;
       	Instr2 <= Instr2_in;
       	Instr3 <= Instr3_in;
       	
        
       	
       	DataFromDPU0<=DataFromDPU0_in;
       	DataFromDPU1<=DataFromDPU1_in;
       	DataFromDPU2<=DataFromDPU2_in;
       	DataFromDPU3<=DataFromDPU3_in;
       	
       	case inst_counter is 
	       	when "00" =>   InstrReg_out <= Instr_In;   
	       	when "01" =>   InstrReg_out2 <= Instr_In;  
	       	when "10" =>   InstrReg_out3 <= Instr_In;  
	       	when "11" =>   InstrReg_out4 <= Instr_In;  
	       	when others => NULL;	
       	end case;

       	
       State_out1 <= State_out;
       State_out2 <= State_out1;
       State_out3 <= State_out2;
  end if; 
  end process; 
  ---------------------------------------------

  
 process(State_out, rst)begin
 	if rst= '1' then 
 		State_in <=Fetch;
 	else
	 	case State_out is
	 		When Fetch => State_in<=Decode;
	 		When Decode => State_in<=Execution;
	 		When Execution => State_in<=WriteBack;
	 		When WriteBack => State_in<=Fetch;
	 	end case;
 	end if;
 end process;
  
 process(Flush,rst,inst_counter,opcode0,opcode1,opcode2,opcode3,Instr, DPU_Flags, DataFromDPU)begin	
 	if  rst = '1' then 
 		Instr0_in<=NOP;
 		Instr1_in<=NOP;
 		Instr2_in<=NOP;
 		Instr3_in<=NOP;
 		DataFromDPU0_in <= (others=> '0');
 		DataFromDPU1_in <= (others=> '0'); 
 		DataFromDPU2_in <= (others=> '0');
 		DataFromDPU3_in <= (others=> '0');
 		DPU_Flags0  <= (others=> '0');
 		DPU_Flags1  <= (others=> '0');
 		DPU_Flags2  <= (others=> '0');
 		DPU_Flags3  <= (others=> '0');
 		opcode <= (others => '0');
 	elsif Flush = '1' then
 		Instr0_in<=HALT;
 		Instr1_in<=HALT;
 		Instr2_in<=HALT;
 		Instr3_in<=HALT;
 		opcode <= (others => '0');
 	else
 		case inst_counter is 
 			when "00" =>    opcode <= opcode3; Instr0_in<=Instr; DataFromDPU0_in<=DataFromDPU; DPU_Flags0  <= DPU_Flags;
	       	when "01" =>    opcode <= opcode0; Instr1_in<=Instr; DataFromDPU1_in<=DataFromDPU; DPU_Flags1  <= DPU_Flags;
	       	when "10" =>    opcode <= opcode1; Instr2_in<=Instr; DataFromDPU2_in<=DataFromDPU; DPU_Flags2  <= DPU_Flags;
	       	when "11" =>    opcode <= opcode2; Instr3_in<=Instr; DataFromDPU3_in<=DataFromDPU; DPU_Flags3  <= DPU_Flags;
	       	when others => NULL;	
       	end case;
       	end if;
end process;

process(rst, Instr0,Instr1,Instr2,Instr3,State_out,State_out1,State_out2,State_out3)begin	
	if rst= '1' then 
 		InstrForExecute_in <= NOP;
 	else
 			if State_out = Execution then 
 				InstrForExecute_in <= Instr0;
 			elsif State_out1 = Execution then
 				InstrForExecute_in <= Instr1;
 			elsif State_out2 = Execution then
 				InstrForExecute_in <= Instr2;
 			elsif State_out3 = Execution then
 				InstrForExecute_in <= Instr3;
 			else
 				InstrForExecute_in <= NOP;
 			end if;
 	end if; 
end process;

process( rst,Instr0,Instr1,Instr2,Instr3,State_out,State_out1,State_out2,State_out3)begin	
	if rst= '1' then 
 		InstrForWriteBack <= NOP;
 	else
 			if State_out = WriteBack then 
 				InstrForWriteBack <= Instr0;
 			elsif State_out1 = WriteBack then
 				InstrForWriteBack <= Instr1;
 			elsif State_out2 = WriteBack then
 				InstrForWriteBack <= Instr2;
 			elsif State_out3 = WriteBack then
 				InstrForWriteBack <= Instr3;
 			else
 				InstrForWriteBack <= NOP;
 			end if;
 	end if; 
end process;

process(InstrForExecute, DataFromDPU, InstrReg_out, InstrReg_out(2 downto 0), InstrReg_out(7 downto 0), PC_out, SP_out)begin
	
	DataToDPU <= (others => '0');
    CommandToDPU <= "00000001000"; --do not do anything
	Reg_in_sel<="00000000";
	Reg_out_sel<="000";
	
	if InstrForExecute = Add_A_R then 
		CommandToDPU <= "00000000010"; 
		
		if State_out = Execution then 
			Reg_out_sel<= InstrReg_out3 (2 downto 0);
		elsif State_out1 = Execution then
			Reg_out_sel<= InstrReg_out4 (2 downto 0);
		elsif State_out2 = Execution then
			Reg_out_sel<= InstrReg_out  (2 downto 0);
		elsif State_out3 = Execution then
			Reg_out_sel<= InstrReg_out2 (2 downto 0);
		else
			Reg_out_sel<= (others => '0');
		end if;  
		
		
	elsif InstrForExecute = Add_A_Mem then 
		if State_out = Execution then 
			MemRdAddress <= InstrReg_out3 (BitWidth-1 downto 0);
		elsif State_out1 = Execution then
			MemRdAddress <= InstrReg_out4 (BitWidth-1 downto 0);
		elsif State_out2 = Execution then
			MemRdAddress <= InstrReg_out (BitWidth-1 downto 0);
		elsif State_out3 = Execution then
			MemRdAddress <= InstrReg_out2 (BitWidth-1 downto 0);
		else
			MemRdAddress<= (others => '0');
		end if;  
		
       	CommandToDPU <= "00000000000";
     elsif InstrForExecute = Add_A_Dir then 
     	if State_out = Execution then 
			DataToDPU <= InstrReg_out3 (BitWidth-1 downto 0);
		elsif State_out1 = Execution then
			DataToDPU <= InstrReg_out4 (BitWidth-1 downto 0);
		elsif State_out2 = Execution then
			DataToDPU <= InstrReg_out (BitWidth-1 downto 0);
		elsif State_out3 = Execution then
			DataToDPU <= InstrReg_out2 (BitWidth-1 downto 0);
		else
			DataToDPU<= (others => '0');
		end if;
        	
       	CommandToDPU <= "00000000001";
     -----------------------------------------------
     elsif InstrForExecute = Sub_A_R then 
       	CommandToDPU <= "00000000110";
		if State_out = Execution then 
			Reg_out_sel<= InstrReg_out3 (2 downto 0);
		elsif State_out1 = Execution then
			Reg_out_sel<= InstrReg_out4 (2 downto 0);
		elsif State_out2 = Execution then
			Reg_out_sel<= InstrReg_out (2 downto 0);
		elsif State_out3 = Execution then
			Reg_out_sel<= InstrReg_out2 (2 downto 0);
		else
			Reg_out_sel<= (others => '0');
		end if;  
     elsif InstrForExecute = Sub_A_Mem then 
       	if State_out = Execution then 
			MemRdAddress <= InstrReg_out3 (BitWidth-1 downto 0);
		elsif State_out1 = Execution then
			MemRdAddress <= InstrReg_out4 (BitWidth-1 downto 0);
		elsif State_out2 = Execution then
			MemRdAddress <= InstrReg_out (BitWidth-1 downto 0);
		elsif State_out3 = Execution then
			MemRdAddress <= InstrReg_out2 (BitWidth-1 downto 0);
		else
			MemRdAddress<= (others => '0');
		end if;
       	CommandToDPU <= "00000000100";
     elsif InstrForExecute = Sub_A_Dir then 
       if State_out = Execution then 
			DataToDPU <= InstrReg_out3 (BitWidth-1 downto 0);
		elsif State_out1 = Execution then
			DataToDPU <= InstrReg_out4 (BitWidth-1 downto 0);
		elsif State_out2 = Execution then
			DataToDPU <= InstrReg_out (BitWidth-1 downto 0);
		elsif State_out3 = Execution then
			DataToDPU <= InstrReg_out2 (BitWidth-1 downto 0);
		else
			DataToDPU<= (others => '0');
		end if;
       	CommandToDPU <= "00000000101";
     -----------------------------------------------                  
     elsif InstrForExecute = IncA then 
       	CommandToDPU <= "00000000011";
     elsif InstrForExecute = DecA then 
       	CommandToDPU <= "00000000111";
     -----------------------------------------------------------
     -----------------------Shift-------------------------------
     -----------------------------------------------------------
     elsif InstrForExecute = ShiftA_R  then
       	CommandToDPU <= "00000011100";
     elsif InstrForExecute = ShiftA_L  then
       	CommandToDPU <= "00000100000";
     elsif InstrForExecute = ShiftArithL  then
       	CommandToDPU <= "00000101100";
     elsif InstrForExecute = ShiftArithR  then
       	CommandToDPU <= "00000101000";
     elsif InstrForExecute = RRC  then
       	CommandToDPU <= "00000111000";  
     elsif InstrForExecute = RLC  then
       	CommandToDPU <= "00000111100";  
     -----------------------------------------------------------
     -----------------------Logical-----------------------------
     -----------------------------------------------------------
      elsif InstrForExecute = NegA then 
       	CommandToDPU <= "00000100100";
      elsif InstrForExecute = FlipA then 
       	CommandToDPU <= "00000110000";
     elsif InstrForExecute = And_A_R  then
       	CommandToDPU <= "00000010010";
		if State_out = Execution then 
			Reg_out_sel<= InstrReg_out3 (2 downto 0);
		elsif State_out1 = Execution then
			Reg_out_sel<= InstrReg_out4 (2 downto 0);
		elsif State_out2 = Execution then
			Reg_out_sel<= InstrReg_out (2 downto 0);
		elsif State_out3 = Execution then
			Reg_out_sel<= InstrReg_out2 (2 downto 0);
		else
			Reg_out_sel<= (others => '0');
		end if; 
     elsif InstrForExecute = OR_A_R  then
       	CommandToDPU <= "00000010110";
		if State_out = Execution then 
			Reg_out_sel<= InstrReg_out3 (2 downto 0);
		elsif State_out1 = Execution then
			Reg_out_sel<= InstrReg_out4 (2 downto 0);
		elsif State_out2 = Execution then
			Reg_out_sel<= InstrReg_out (2 downto 0);
		elsif State_out3 = Execution then
			Reg_out_sel<= InstrReg_out2 (2 downto 0);
		else
			Reg_out_sel<= (others => '0');
		end if; 
     elsif InstrForExecute = XOR_A_R  then
       	CommandToDPU <= "00000011010";
		if State_out = Execution then 
			Reg_out_sel<= InstrReg_out3 (2 downto 0);
		elsif State_out1 = Execution then
			Reg_out_sel<= InstrReg_out4 (2 downto 0);
		elsif State_out2 = Execution then
			Reg_out_sel<= InstrReg_out (2 downto 0);
		elsif State_out3 = Execution then
			Reg_out_sel<= InstrReg_out2 (2 downto 0);
		else
			Reg_out_sel<= (others => '0');
		end if; 
     -----------------------------------------------------------
     -----------------------Memory------------------------------
     -----------------------------------------------------------
     elsif InstrForExecute = Load_R0_Mem  then
       	if State_out = Execution then 
			MemRdAddress <= InstrReg_out3 (BitWidth-1 downto 0);
		elsif State_out1 = Execution then
			MemRdAddress <= InstrReg_out4 (BitWidth-1 downto 0);
		elsif State_out2 = Execution then
			MemRdAddress <= InstrReg_out (BitWidth-1 downto 0);
		elsif State_out3 = Execution then
			MemRdAddress <= InstrReg_out2 (BitWidth-1 downto 0);
		else
			MemRdAddress<= (others => '0');
		end if;
        	CommandToDPU <= "11000001000";
       	Reg_in_sel<= "00000001";
     elsif InstrForExecute = Load_A_Mem  then
       	if State_out = Execution then 
			MemRdAddress <= InstrReg_out3 (BitWidth-1 downto 0);
		elsif State_out1 = Execution then
			MemRdAddress <= InstrReg_out4 (BitWidth-1 downto 0);
		elsif State_out2 = Execution then
			MemRdAddress <= InstrReg_out (BitWidth-1 downto 0);
		elsif State_out3 = Execution then
			MemRdAddress <= InstrReg_out2 (BitWidth-1 downto 0);
		else
			MemRdAddress<= (others => '0');
		end if;
       	CommandToDPU <= "00000001100";
     elsif InstrForExecute = SavePC  then
       	DataToDPU <= PC_out;
       	CommandToDPU <= "00000001101";	  
     elsif InstrForExecute = Load_R0_Dir  then
         CommandToDPU <= "01000001000";
        if State_out = Execution then 
			DataToDPU <= InstrReg_out3 (BitWidth-1 downto 0);
		elsif State_out1 = Execution then
			DataToDPU <= InstrReg_out4 (BitWidth-1 downto 0);
		elsif State_out2 = Execution then
			DataToDPU <= InstrReg_out (BitWidth-1 downto 0);
		elsif State_out3 = Execution then
			DataToDPU <= InstrReg_out2 (BitWidth-1 downto 0);
		else
			DataToDPU<= (others => '0');
		end if;
         Reg_in_sel<= "00000001";
		 Reg_out_sel<= "000";
		 
     elsif InstrForExecute = Load_Ind_A  then   
     	MemRdAddress <= DataFromDPU;
        CommandToDPU <= "00000001100";
        Reg_in_sel<= "00000000";
		Reg_out_sel<= "000";
		
     elsif InstrForExecute = load_A_R then 
         CommandToDPU <= "00000001100";
		if State_out = Execution then 
			Reg_out_sel<= InstrReg_out3 (2 downto 0);
		elsif State_out1 = Execution then
			Reg_out_sel<= InstrReg_out4 (2 downto 0);
		elsif State_out2 = Execution then
			Reg_out_sel<= InstrReg_out (2 downto 0);
		elsif State_out3 = Execution then
			Reg_out_sel<= InstrReg_out2 (2 downto 0);
		else
			Reg_out_sel<= (others => '0');
		end if; 
		
     elsif InstrForExecute = load_R_A then 
     	CommandToDPU <= "10000001000";    
     	if State_out = Execution then 
			Reg_in_sel<= InstrReg_out3 (BitWidth-1 downto 0);
		elsif State_out1 = Execution then
			Reg_in_sel<= InstrReg_out4 (BitWidth-1 downto 0);
		elsif State_out2 = Execution then
			Reg_in_sel<= InstrReg_out (BitWidth-1 downto 0);
		elsif State_out3 = Execution then
			Reg_in_sel<= InstrReg_out2 (BitWidth-1 downto 0);
		else
			Reg_in_sel<= (others => '0');
		end if;  
     -----------------------------------------------------------
     -----------------------Stack-------------------------------
     -----------------------------------------------------------
     elsif InstrForExecute = POP  then
         MemRdAddress <=   SP_out - "00000001";
         CommandToDPU <= "00000001100";
     -----------------------------------------------------------
     -----------------------ClearFlags--------------------------
     -----------------------------------------------------------
     elsif InstrForExecute = ClearZ  then
         CommandToDPU <= "00001001000";
     elsif InstrForExecute = ClearOV  then
        	CommandToDPU <= "00010001000"; 
     elsif InstrForExecute = ClearC  then    
        	CommandToDPU <= "00100001000";
     elsif InstrForExecute = ClearACC  then    
        	CommandToDPU <= "00000110100";   
     else
     	CommandToDPU <= "00000001000"; --do not do anything
     end if;

end process;
          
   
process(InstrForWriteBack,InstrReg_out,SP_out,PC_out, InstrReg_out2, InstrReg_out3, InstrReg_out4, State_out, State_out1, State_out2, State_out3, DPU_Flags0(0), DPU_Flags0(1), DPU_Flags0(2), DPU_Flags0(3), DPU_Flags1(0), DPU_Flags1(1), DPU_Flags1(2), DPU_Flags1(3), DPU_Flags2(0), DPU_Flags2(1), DPU_Flags2(2), DPU_Flags2(3), DPU_Flags3(0), DPU_Flags3(1), DPU_Flags3(2), DPU_Flags3(3), DataFromDPU0, DataFromDPU1, DataFromDPU2, DataFromDPU3) 
	begin
	SP_in <= SP_out;
	Mem_RW <= '0';
	MemWrtAddress <= (others => '0');
	Flush<='0';
  if InstrForWriteBack = Store_A_Mem then 
        MemWrtAddress <= InstrReg_out (BitWidth-1 downto 0);
        Mem_RW <= '1';
      	PC_in <= PC_out+1;	
  elsif InstrForWriteBack = HALT then 
      	PC_in <= PC_out;
      	Flush<='1';
      -----------------------------------------------------------
      -----------------------Jump--------------------------------
      -----------------------------------------------------------      
  	elsif InstrForWriteBack = Jmp then 
  	
		if State_out = WriteBack then 
			PC_in <= InstrReg_out4 (BitWidth-1 downto 0);
		elsif State_out1 = WriteBack then
			PC_in <= InstrReg_out (BitWidth-1 downto 0);
		elsif State_out2 = WriteBack then
			PC_in <= InstrReg_out2 (BitWidth-1 downto 0);
		elsif State_out3 = WriteBack then
			PC_in <= InstrReg_out3 (BitWidth-1 downto 0);
		else
			PC_in <= (others => '0');
		end if;  
        
	elsif InstrForWriteBack = JmpOV then    
		
	 	if State_out = WriteBack then 
			if DPU_Flags3(3) = '1' then
				PC_in <= InstrReg_out4 (BitWidth-1 downto 0);
			else
				PC_in <= PC_out+1;
			end if;
		elsif State_out1 = WriteBack then
			if DPU_Flags0(3) = '1' then
				PC_in <= InstrReg_out (BitWidth-1 downto 0);
			else
				PC_in <= PC_out+1;
			end if;
		elsif State_out2 = WriteBack then
			if DPU_Flags1(3) = '1' then
				PC_in <= InstrReg_out2  (BitWidth-1 downto 0);
			else
				PC_in <= PC_out+1;
			end if;
		elsif State_out3 = WriteBack then
			if DPU_Flags2(3) = '1' then
				PC_in <= InstrReg_out3 (BitWidth-1 downto 0);
			else
				PC_in <= PC_out+1;
			end if;
		else
			PC_in <= (others => '0');
		end if;  
	
   	elsif InstrForWriteBack = JmpZ  then   
         	
		if State_out = WriteBack then 
			if DPU_Flags3(3) = '1' then
				PC_in <= InstrReg_out4 (BitWidth-1 downto 0);
			else
				PC_in <= PC_out+1;
			end if;
		elsif State_out1 = WriteBack then
			if DPU_Flags0(3) = '1' then
				PC_in <= InstrReg_out (BitWidth-1 downto 0);
			else
				PC_in <= PC_out+1;
			end if;
		elsif State_out2 = WriteBack then
			if DPU_Flags1(3) = '1' then
				PC_in <= InstrReg_out2  (BitWidth-1 downto 0);
			else
				PC_in <= PC_out+1;
			end if;
		elsif State_out3 = WriteBack then
			if DPU_Flags2(3) = '1' then
				PC_in <= InstrReg_out3 (BitWidth-1 downto 0);
			else
				PC_in <= PC_out+1;
			end if;
		else
			PC_in <= (others => '0');
		end if; 
		
   	elsif InstrForWriteBack = JMPEQ  then   
         	
	 	if State_out = WriteBack then 
			if DPU_Flags3(3) = '1' then
				PC_in <= InstrReg_out4 (BitWidth-1 downto 0);
			else
				PC_in <= PC_out+1;
			end if;
		elsif State_out1 = WriteBack then
			if DPU_Flags0(3) = '1' then
				PC_in <= InstrReg_out (BitWidth-1 downto 0);
			else
				PC_in <= PC_out+1;
			end if;
		elsif State_out2 = WriteBack then
			if DPU_Flags1(3) = '1' then
				PC_in <= InstrReg_out2  (BitWidth-1 downto 0);
			else
				PC_in <= PC_out+1;
			end if;
		elsif State_out3 = WriteBack then
			if DPU_Flags2(3) = '1' then
				PC_in <= InstrReg_out3 (BitWidth-1 downto 0);
			else
				PC_in <= PC_out+1;
			end if;
		else
			PC_in <= (others => '0');
		end if; 
		
   	elsif InstrForWriteBack = JmpC   then    
         	
	 	if State_out = WriteBack then 
			if DPU_Flags3(3) = '1' then
				PC_in <= InstrReg_out4 (BitWidth-1 downto 0);
			else
				PC_in <= PC_out+1;
			end if;
		elsif State_out1 = WriteBack then
			if DPU_Flags0(3) = '1' then
				PC_in <= InstrReg_out (BitWidth-1 downto 0);
			else
				PC_in <= PC_out+1;
			end if;
		elsif State_out2 = WriteBack then
			if DPU_Flags1(3) = '1' then
				PC_in <= InstrReg_out2  (BitWidth-1 downto 0);
			else
				PC_in <= PC_out+1;
			end if;
		elsif State_out3 = WriteBack then
			if DPU_Flags2(3) = '1' then
				PC_in <= InstrReg_out3 (BitWidth-1 downto 0);
			else
				PC_in <= PC_out+1;
			end if;
		else
			PC_in <= (others => '0');
		end if;
		
	elsif InstrForWriteBack= Jmp_rel then 
			
		if State_out = WriteBack then 
			PC_in <= PC_out + InstrReg_out4 (BitWidth-1 downto 0);
		elsif State_out1 = WriteBack then
			PC_in <= PC_out + InstrReg_out  (BitWidth-1 downto 0);
		elsif State_out2 = WriteBack then
			PC_in <= PC_out + InstrReg_out2  (BitWidth-1 downto 0);
		elsif State_out3 = WriteBack then
			PC_in <= PC_out + InstrReg_out3 (BitWidth-1 downto 0);
		else
			PC_in <= (others => '0');
		end if;  
        
	elsif InstrForWriteBack= LoadPC then
		
		if State_out = WriteBack then 
			 PC_in <= DataFromDPU2;
		elsif State_out1 = WriteBack then
			 PC_in <= DataFromDPU3;
		elsif State_out2 = WriteBack then
			 PC_in <= DataFromDPU0;
		elsif State_out3 = WriteBack then
			 PC_in <= DataFromDPU1;
		else
			PC_in <= (others => '0');
		end if;  
		
 
   	-----------------------------------------------------------
   	-----------------------Stack OP----------------------------
   	-----------------------------------------------------------
   	elsif InstrForWriteBack= PUSH then
  		 PC_in <= PC_out+1;
         MemWrtAddress <=  SP_out;
         SP_in <= SP_out + 1;
         Mem_RW <= '1'; 
    elsif InstrForWriteBack = POP  then
         SP_in <= SP_out - 1;
   	else
   		PC_in <= PC_out+1;
   	end if;    
      
end process;
  
  
  
  
------------------------------------------------
-- Instr decoder
------------------------------------------------
process (opcode) 
  begin
    case opcode is
      when "000000" => Instr <= Add_A_R;
      when "000001" => Instr <= Add_A_Mem;
      when "000010" => Instr <= Add_A_Dir;
         
      when "000011" => Instr <= Sub_A_R;   
      when "000100" => Instr <= Sub_A_Mem;  
      when "000101" => Instr <= Sub_A_Dir; 
        
      when "000110" => Instr <= IncA; 
      when "000111" => Instr <= DecA; 
      
        
      when "001000" => Instr <= ShiftArithR;
      when "001001" => Instr <= ShiftArithL;      
      when "001010" => Instr <= ShiftA_R; 
      when "001011" => Instr <= ShiftA_L; 
      when "001100" => Instr <= RRC;        
      when "001101" => Instr <= RLC; 
        
      when "001110" => Instr <= And_A_R; 
      when "001111" => Instr <= OR_A_R;
      when "010000" => Instr <= XOR_A_R;
      when "010001" => Instr <= FlipA;
      when "010010" => Instr <= NegA; 

      when "010011" => Instr <= Jmp;  
      when "010100" => Instr <= JmpZ;
      when "010101" => Instr <= JmpOV;
      when "010110" => Instr <= JmpC;  
      when "010111" => Instr <= Jmp_rel;
      when "011000" => Instr <= JMPEQ;
        
      when "011001" => Instr <= ClearZ;
      when "011010" => Instr <= ClearOV;
      when "011011" => Instr <= ClearC; 
      when "011100" => Instr <= ClearACC;   
        
      when "011101" => Instr <= LoadPC;
      when "011110" => Instr <= SavePC;              
      when "011111" => Instr <= Load_A_Mem;
      when "100000" => Instr <= Store_A_Mem;  
      when "100001" => Instr <= Load_R0_Dir;  
      when "100010" => Instr <= Load_R0_Mem;  
     
      when  "100011" => Instr <= load_A_R;
      when  "100100" => Instr <= load_R_A;
      when  "100101" => Instr <= Load_Ind_A ;
        
      when "111100" => Instr <= PUSH;
      when "111101" => Instr <= POP;
      
      when "111110" => Instr <= NOP;
      when "111111" => Instr <= HALT; 

      when others =>  Instr <= NOP;
      end case;     
end process;
  
Instr_Add <= PC_out; 
  
end RTL;
